/*
 *  HCI UART driver initialization utility for Marvell 8688 chipset
 *  Borqs for Spark
 *  Liao Cheng <cheng.liao@borqs.com>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <termios.h>
#include <time.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <sys/uio.h>
#include <pthread.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include <hardware/wifi.h>

#ifndef N_TTY
#define N_TTY	0
#endif

#ifndef N_HCI
#define N_HCI	15
#endif

#define HCIUARTSETPROTO		_IOW('U', 200, int)

#define HCI_UART_H4	0
#define HCI_UART_BCSP	1

struct uart_t {
	char *type;
	int  m_id;
	int  p_id;
	int  proto;
	int  flags;
	char *bdaddr;
	int  (*init) (int fd, struct uart_t *u, struct termios *ti);
};

#define FLOW_CTL	0x0001

static volatile sig_atomic_t __io_canceled = 0;
static int warn_reset = 0;
static void *hciattach_thread(void *argv);
static int uart_fd = 0;

static void sig_hup(int sig)
{
}

static void sig_term(int sig)
{
	__io_canceled = 1;
}

static void sig_alarm(int sig)
{
	fprintf(stderr, "Initialization timed out.\n");
	exit(1);
}

static int set_speed(int fd, struct termios *ti)
{
	if(warn_reset)
	{
		printf("Set baud rate to 1500000.\n");
		cfsetospeed(ti, B1500000);
		cfsetispeed(ti, B1500000);
	}
	else
	{
		printf("Set baud rate to 38400.\n");
		cfsetospeed(ti, B38400);
		cfsetispeed(ti, B38400);
	}

	return tcsetattr(fd, TCSANOW, ti);
}

static int read_check(int fd, void *buf, int count)
{
	int res;

	do {
		res = read(fd, buf, count);
		if (res != -1) {
			buf += res; 
			count -= res;
		}
	} while (count && (errno == 0 || errno == EINTR));

	if (count)
		return -1;

	return 0;
}

struct uart_t uart[] = {
	{ "any",        0x0000, 0x0000, HCI_UART_H4,	FLOW_CTL, NULL, NULL    },
	{ "bcsp",       0x0000, 0x0000, HCI_UART_BCSP, 	FLOW_CTL,        NULL, NULL    },
	{ NULL, 	0, 	0, 	0, 		0, 	NULL, 	NULL 	}
};

struct uart_t * get_by_id(int m_id, int p_id)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (uart[i].m_id == m_id && uart[i].p_id == p_id)
			return &uart[i];
	}
	return NULL;
}

struct uart_t * get_by_type(char *type)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (!strcmp(uart[i].type, type))
			return &uart[i];
	}
	return NULL;
}

int reset_uart(int fd)
{
	struct termios ti;

	if (tcgetattr(fd, &ti) < 0) {
		printf("Can't get port settings.\n");
		return -1;
	}
	printf("reset the uart baud rate to 38400.\n");

	cfsetospeed(&ti, B38400);
	cfsetispeed(&ti, B38400);

	return tcsetattr(fd, TCSANOW, &ti);
}

/* Initialize UART driver */
int init_uart(char *dev, struct uart_t *u, int send_break)
{
	struct termios ti;
	int fd, i;

	fd = open(dev, O_RDWR | O_NOCTTY);
	tcflush(fd, TCIOFLUSH);
	if (fd < 0) {
		perror("Can't open serial port");
		return -1;
	}

	if (tcgetattr(fd, &ti) < 0) {
		perror("Can't get port settings");
		return -1;
	}

	cfmakeraw(&ti);

	ti.c_cflag |= CLOCAL;
	if (u->flags & FLOW_CTL)
		ti.c_cflag |= CRTSCTS;
	else
		ti.c_cflag &= ~CRTSCTS;

	if (tcsetattr(fd, TCSANOW, &ti) < 0) {
		perror("Can't set port settings");
		return -1;
	}

	/* Set initial baudrate */
	if (set_speed(fd, &ti) < 0) {
		perror("Can't set initial baud rate");
		return -1;
	}

	tcflush(fd, TCIOFLUSH);

	if (send_break) {
		tcsendbreak(fd, 0);
		usleep(500000);
	}

	if (u->init && u->init(fd, u, &ti) < 0)
		return -1;

	tcflush(fd, TCIOFLUSH);

	if (set_speed(fd, &ti) < 0) {
		perror("Can't set baud rate");
		return -1;
	}

	i = N_HCI;
	if (ioctl(fd, TIOCSETD, &i) < 0) {
		perror("Can't set line discipline");
		return -1;
	}

	if (ioctl(fd, HCIUARTSETPROTO, u->proto) < 0) {
		perror("Can't set device");
		return -1;
	}

	return fd;
}

int main(int argc, char *argv[])
{
	struct uart_t *u = NULL;
	int detach, opt, i, n, ld, err;
	int to = 3;
	int send_break = 0;
	pid_t pid;
	struct sigaction sa;
	struct pollfd p;
	char dev[PATH_MAX];

	int number = 0;
	int flag = 0;
	pthread_t hciattach_pid;

	detach = 1;
	warn_reset = 0;

	wifi_load_driver();
	usleep(3000000);

	while ((opt=getopt(argc, argv, "mbnpt:s")) != EOF) {
		switch(opt) {
		case 'm':
			flag = 1;
			break;

		case 'b':
			send_break = 1;
			break;

		case 'n':
			detach = 0;
			break;

		case 't':
			to = atoi(optarg);
			break;

		default:
			exit(1);
		}
	}

	n = argc - optind;
	if (n < 2) {
		exit(1);
	}
CSR:
	for (n = 0; optind < argc; n++, optind++) {
		char *opt;

		opt = argv[optind];

		switch(n) {
		case 0:
			dev[0] = 0;
			if (!strchr(opt, '/'))
				strcpy(dev, "/dev/");
			strcat(dev, opt);
			break;

		case 1:
			u = get_by_type("any"); // for Marvell 8688

			if (!u) {
				fprintf(stderr, "Unknown device type or id\n");
				exit(1);
			}
			
			break;

		case 2:
			break;

		case 3:
			u->flags |=  FLOW_CTL; // Marvell 8688 need add UART flow control
			break;

		case 4:
			u->bdaddr = argv[optind];
			break;
		}
	}

	if (!u) {
		fprintf(stderr, "Unknown device type or id\n");
		exit(1);
	}

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = sig_alarm;
	sigaction(SIGALRM, &sa, NULL);

	/* 5 seconds should be enough for initialization */
	alarm(to);

	uart_fd = init_uart(dev, u, send_break);

	if (uart_fd < 0) {
		perror("Can't initialize device"); 
		exit(1);
	}

	alarm(0);

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);

	sa.sa_handler = sig_hup;
	sigaction(SIGHUP, &sa, NULL);

	p.fd = uart_fd;
	p.events = POLLERR | POLLHUP;

	printf("Bluetooth 8688: start enter main loop... \n");
	while (!__io_canceled) {	
		p.revents = 0;
		err = poll(&p, 1, 500);
		if (err < 0 && errno == EINTR)
			continue;
		if (err)
			break;

		if(flag)
		{
			if(number == 2)
			{
				number = 30;
				pthread_create(&hciattach_pid, NULL, hciattach_thread, "A");
			}
			else if((number > 30) && warn_reset)
			{
				__io_canceled = 1;
			}
			number++;
		}
	}

	printf("BCSP: exit main loop... \n");

	/* Restore TTY line discipline */
	ld = N_TTY;
	if (ioctl(uart_fd, TIOCSETD, &ld) < 0) {
		perror("Can't restore line discipline");
	}

	if(flag)
	{
		close(uart_fd);
		flag = 0;
		__io_canceled = 0;
		goto CSR;
	}
	reset_uart(uart_fd);
	usleep(100000);
	close(uart_fd);

	wifi_unload_driver();

	return 0;
}

static void cmd_rstat(int ctl, int hdev, char *opt)
{
	/* Reset HCI device stat counters */
	if (ioctl(ctl, HCIDEVRESTAT, hdev) < 0) {
		fprintf(stderr, "Can't reset stats counters hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_up(int ctl, int hdev, char *opt)
{
        /* Start HCI device */
       	if (ioctl(ctl, HCIDEVUP, hdev) < 0) {
       	        if (errno == EALREADY)
       	                return;
       	        fprintf(stderr, "Can't init device hci%d: %s (%d)\n",
       	                                        hdev, strerror(errno), errno);
	      	exit(1);
	}
}

static void cmd_down(int ctl, int hdev, char *opt)
{
	/* Stop HCI device */
	if (ioctl(ctl, HCIDEVDOWN, hdev) < 0) {
		fprintf(stderr, "Can't down device hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_reset(int ctl, int hdev, char *opt)
{
	/* Reset HCI device */
	cmd_down(ctl, hdev, "down");
	cmd_up(ctl, hdev, "up");
}

static struct {
	char *cmd;
	void (*func)(int ctl, int hdev, char *opt);
	char *opt;
	char *doc;
} command[] = {
	{ "up",		cmd_up,		0,		"Open and initialize HCI device" },
	{ "down",	cmd_down,	0,		"Close HCI device" },
	{ "reset",	cmd_reset,	0,		"Reset HCI device" },
	{ "rstat",	cmd_rstat,	0,		"Reset statistic counters" },
	{ NULL, NULL, 0, NULL }
};

static void *hciattach_thread(void *argv)
{
	int dd;
	unsigned char cp[254];

	printf("LIAO: HCI dev up start......\n");
	usleep(8000000);

	dd = hci_open_dev(0);

	memset(&cp, 0, sizeof(cp));
	cp[0] = 0x04;

	if (hci_send_cmd(dd, 0x3f, 0x0007, 1, cp) < 0) {
		printf("Send Command HCI_CMD_MARVELL_WRITE_PCM_SETTINGS failed.\n");
	}
	usleep(50000);

	memset(&cp, 0, sizeof(cp));
	cp[0] = 0x00;
	cp[1] = 0x20;
	cp[2] = 0x1c;
	cp[3] = 0x00;

	if (hci_send_cmd(dd, 0x3f, 0x0009, 4, cp) < 0) {
		printf("Send failed\n");
	}
	usleep(50000);

	printf("LIAO: HCI dev up end......\n");
	hci_close_dev(dd);

	usleep(1000000);
	warn_reset = 1;
	pthread_exit(NULL);

	return 0;
}

