/*
 *  HCI UART driver initialization utility for CSR BC4 chipset
 *  Borqs for Jadespring
 *  Liao Cheng <cheng.liao@borqs.com>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <termios.h>
#include <time.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
//#include <sys/endian.h>
#include <sys/uio.h>
#include <pthread.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include <arpa/inet.h>
//#include <netdb.h>
#include <netinet/in.h>

#include "csr.h"

#ifndef DEFFILEMODE
#define DEFFILEMODE 0
#endif
#define PATH_FILE_UART "/sdcard/bluetooth/"
#define FILE_UART_TEST	"/sdcard/bluetooth/.uart_test"

#define TTYS0_DEV "/dev/ttyS0"
#define BT_DEV "/dev/ttyS1"
#define SERVER_ADDRESS "192.168.100.1"
#define SERVER_PORT 1500

#define BUFF_LEN 80

static int sock_fd;
int fd_test_uart;

#ifndef N_TTY
#define N_TTY	0
#endif

#ifndef N_HCI
#define N_HCI	15
#endif

#define HCIUARTSETPROTO		_IOW('U', 200, int)

#define HCI_UART_H4	0
#define HCI_UART_BCSP	1

struct uart_t {
	char *type;
	int  m_id;
	int  p_id;
	int  proto;
	int  flags;
	char *bdaddr;
	int  (*init) (int fd, struct uart_t *u, struct termios *ti);
};

#define FLOW_CTL	0x0001

static volatile sig_atomic_t __io_canceled = 0;
static int exit_flag = 0;
static int warn_reset = 0;
static void *hciattach_thread(void *argv);
static int uart_fd = 0;

static void sig_hup(int sig)
{
}

static void sig_term(int sig)
{
	__io_canceled = 1;
}

static void sig_alarm(int sig)
{
	fprintf(stderr, "Initialization timed out.\n");
	exit(1);
}

static int set_speed(int fd, struct termios *ti)
{
	if(warn_reset)
	{
		cfsetospeed(ti, B1500000);
		cfsetispeed(ti, B1500000);
	}
	else
	{
		cfsetospeed(ti, B921600);
		cfsetispeed(ti, B921600);
	}

	return tcsetattr(fd, TCSANOW, ti);
}

static int read_check(int fd, void *buf, int count)
{
	int res;

	do {
		res = read(fd, buf, count);
		if (res != -1) {
			buf += res; 
			count -= res;
		}
	} while (count && (errno == 0 || errno == EINTR));

	if (count)
		return -1;

	return 0;
}

/*
 * BCSP specific initialization
 */
int serial_fd;

static void bcsp_tshy_sig_alarm(int sig)
{
	static int retries=0;
	unsigned char bcsp_sync_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xda,0xdc,0xed,0xed,0xc0};
	int len;

	if (retries < 20) {
		retries++;
		len = write(serial_fd, &bcsp_sync_pkt, 10);
		alarm(1);
		return;
	}

	tcflush(serial_fd, TCIOFLUSH);
	fprintf(stderr, "BCSP initialization timed out\n");
	exit(1);
}

static void bcsp_tconf_sig_alarm(int sig)
{
	static int retries=0;
	unsigned char bcsp_conf_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xad,0xef,0xac,0xed,0xc0};
	int len;

	if (retries < 10){
		retries++;
		len = write(serial_fd, &bcsp_conf_pkt, 10);
		alarm(1);
		return;
	}

	tcflush(serial_fd, TCIOFLUSH);
	fprintf(stderr, "BCSP initialization timed out\n");
	exit(1);
}

static int bcsp(int fd, struct uart_t *u, struct termios *ti)
{
	unsigned char byte, bcsph[4], bcspp[4],
		bcsp_sync_resp_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xac,0xaf,0xef,0xee,0xc0},
		bcsp_conf_resp_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xde,0xad,0xd0,0xd0,0xc0},
		bcspsync[4]     = {0xda, 0xdc, 0xed, 0xed},
		bcspsyncresp[4] = {0xac,0xaf,0xef,0xee},
		bcspconf[4]     = {0xad,0xef,0xac,0xed},
		bcspconfresp[4] = {0xde,0xad,0xd0,0xd0};
	struct sigaction sa;
	int len;

	printf("hciattach bcsp for CSR chip...\n");
	if (set_speed(fd, ti) < 0) {
		perror("Can't set default baud rate");
		return -1;
	}

	ti->c_cflag |= PARENB;
	ti->c_cflag &= ~(PARODD);

	if (tcsetattr(fd, TCSANOW, ti) < 0) {
		perror("Can't set port settings");
		return -1;
	}

	alarm(0);

	serial_fd = fd;
	memset(&sa, 0, sizeof(sa));
	sa.sa_flags = SA_NOCLDSTOP;
	sa.sa_handler = bcsp_tshy_sig_alarm;
	sigaction(SIGALRM, &sa, NULL);

	/* State = shy */

	bcsp_tshy_sig_alarm(0);
	while (1) {
		do {
			if (read_check(fd, &byte, 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (byte != 0xC0);

		do {
			if ( read_check(fd, &bcsph[0], 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (bcsph[0] == 0xC0);

		if ( read_check(fd, &bcsph[1], 3) == -1){
			perror("Failed to read");
			return -1;
		}

		if (((bcsph[0] + bcsph[1] + bcsph[2]) & 0xFF) != (unsigned char)~bcsph[3])
			continue;
		if (bcsph[1] != 0x41 || bcsph[2] != 0x00)
			continue;

		if (read_check(fd, &bcspp, 4) == -1){
			perror("Failed to read");
			return -1;
		}

		if (!memcmp(bcspp, bcspsync, 4)) {
			len = write(fd, &bcsp_sync_resp_pkt,10);
		} else if (!memcmp(bcspp, bcspsyncresp, 4))
			break;
	}

	/* State = curious */

	alarm(0);
	sa.sa_handler = bcsp_tconf_sig_alarm;
	sigaction(SIGALRM, &sa, NULL);
	alarm(1);

	while (1) {
		do {
			if (read_check(fd, &byte, 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (byte != 0xC0);

		do {
			if (read_check(fd, &bcsph[0], 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (bcsph[0] == 0xC0);

		if (read_check(fd, &bcsph[1], 3) == -1){
			perror("Failed to read");
			return -1;
		}

		if (((bcsph[0] + bcsph[1] + bcsph[2]) & 0xFF) != (unsigned char)~bcsph[3])
			continue;

		if (bcsph[1] != 0x41 || bcsph[2] != 0x00)
			continue;

		if (read_check(fd, &bcspp, 4) == -1){
			perror("Failed to read");
			return -1;
		}

		if (!memcmp(bcspp, bcspsync, 4))
			len = write(fd, &bcsp_sync_resp_pkt, 10);
		else if (!memcmp(bcspp, bcspconf, 4))
			len = write(fd, &bcsp_conf_resp_pkt, 10);
		else if (!memcmp(bcspp, bcspconfresp,  4))
			break;
	}

	/* State = garrulous */
	exit_flag = 1;

	printf("exit bcsp ...\n");
	
	return 0;
}

struct uart_t uart[] = {
	{ "any",        0x0000, 0x0000, HCI_UART_H4,	FLOW_CTL, NULL, NULL    },
	{ "bcsp",       0x0000, 0x0000, HCI_UART_BCSP, 	0,        NULL, bcsp    },
	{ NULL, 	0, 	0, 	0, 		0, 	NULL, 	NULL 	}
};

struct uart_t * get_by_id(int m_id, int p_id)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (uart[i].m_id == m_id && uart[i].p_id == p_id)
			return &uart[i];
	}
	return NULL;
}

struct uart_t * get_by_type(char *type)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (!strcmp(uart[i].type, type))
			return &uart[i];
	}
	return NULL;
}

/* Initialize UART driver */
int init_uart(char *dev, struct uart_t *u, int send_break)
{
	struct termios ti;
	int fd, i;
	
	if(warn_reset){
		fd = uart_fd;
		printf("warn_reset\n");
	}
	else
		fd = open(dev, O_RDWR | O_NOCTTY);
	
	tcflush(fd, TCIOFLUSH);
	if (fd < 0) {
		perror("Can't open serial port");
		return -1;
	}

	if (tcgetattr(fd, &ti) < 0) {
		perror("Can't get port settings");
		return -1;
	}

	cfmakeraw(&ti);

	ti.c_cflag |= CLOCAL;
	if (u->flags & FLOW_CTL)
		ti.c_cflag |= CRTSCTS;
	else
		ti.c_cflag &= ~CRTSCTS;

	if (tcsetattr(fd, TCSANOW, &ti) < 0) {
		perror("Can't set port settings");
		return -1;
	}

	/* Set initial baudrate */
	if (set_speed(fd, &ti) < 0) {
		perror("Can't set initial baud rate");
		return -1;
	}

	tcflush(fd, TCIOFLUSH);

	if (send_break) {
		tcsendbreak(fd, 0);
		usleep(500000);
	}

	if (u->init && u->init(fd, u, &ti) < 0)
		return -1;

	tcflush(fd, TCIOFLUSH);

	if (set_speed(fd, &ti) < 0) {
		perror("Can't set baud rate");
		return -1;
	}

	i = N_HCI;
	if (ioctl(fd, TIOCSETD, &i) < 0) {
		perror("Can't set line discipline");
		return -1;
	}

	if (ioctl(fd, HCIUARTSETPROTO, u->proto) < 0) {
		perror("Can't set device");
		return -1;
	}

	return fd;
}

static int make_socket(int *sockFd, char *address, int port)
{
    int fd;
    struct sockaddr_in name;

    fd = socket(PF_INET, SOCK_STREAM, 0);
    if (fd < 0)
    {
        printf("apgs_client creat socket erorr!\r\n");
        return -1;
    }

    memset(&name, 0, sizeof(struct sockaddr_in));
    name.sin_family = AF_INET;

    if (NULL == address)
    {
        name.sin_addr.s_addr = inet_addr(SERVER_ADDRESS);
    }
    else
    {
        name.sin_addr.s_addr = inet_addr(address);
    }

    if (-1 == port)
    {
        name.sin_port = htons(SERVER_PORT);
    }
    else
    {
        name.sin_port = htons(port);
    }

    if (connect(fd, (struct sockaddr *)&name, sizeof(struct sockaddr_in)) < 0)
    {
        printf("Failed to connect socket\r\n");
        close(fd);
        return -1;
    }

    *sockFd = fd;

    return 0;
}

static void* transfer_loop(void *argv)
{
	int r_c, w_c;
	unsigned char b[BUFF_LEN];
	
	system("hciconfig -a hci0 up &");//power up
	sleep(3);
	system("hcitool cmd 0x03 0x0005 0x02 0x00 0x02 &");//enable events
	sleep(3);
	system("hcitool cmd 0x03 0x001A 0x03 &");//discoverable
	sleep(3);
	system("hcitool cmd 0x06 0x0003 &");//test mode
	sleep(3);
	printf("test mode...\n");

	while (!__io_canceled) {
		if((r_c = read(sock_fd, b, BUFF_LEN)) != 0){
			printf("%s ---> %s\n", TTYS0_DEV, BT_DEV);
			w_c = write(uart_fd, b, r_c);
			printf("w_c = %d, errno = %d\n", w_c, errno);
			write(fd_test_uart, b, r_c);
			printf("%d ---> %d\n", r_c, w_c);
		}
	}

	pthread_exit(NULL);

	return NULL;
}

int main(int argc, char *argv[])
{
	struct uart_t *u = NULL;
	int detach, opt, i, n, ld, err;
	int to = 5;
	int send_break = 0;
	pid_t pid;
	struct sigaction sa;
	struct pollfd p;
	char dev[PATH_MAX];

	unsigned char buff[BUFF_LEN];
	int read_count, write_count;
	pthread_t task;
	pthread_attr_t task_attr;

	int number = 0;
	int flag = 0;
	pthread_t hciattach_pid;

	detach = 1;
	warn_reset = 0;

	const char filepath[] = PATH_FILE_UART;
	const char filename_test[] = FILE_UART_TEST;
	int fd_path, fd_file=-1, port = -1;
	
	while ((opt=getopt(argc, argv, "mbnp:t:s")) != EOF) {
		switch(opt) {
		case 'm':
			flag = 1;
			break;

		case 'b':
			send_break = 1;
			break;

		case 'n':
			detach = 0;
			break;

		case 't':
			to = atoi(optarg);
			break;

		case 'p':
			port = atoi(optarg);
			printf("port = %d\n", port);
			break;

		default:
			printf("Error for getopt. Exit...\n");
			exit(1);
		}
	}

	n = argc - optind;
	if (n < 2) {
		exit(1);
	}
CSR:
	for (n = 0; optind < argc; n++, optind++) {
		char *opt;

		opt = argv[optind];

		switch(n) {
		case 0:
			dev[0] = 0;
			if (!strchr(opt, '/'))
				strcpy(dev, "/dev/");
			strcat(dev, opt);
			break;

		case 1:
			if (strchr(argv[optind], ',')) {
				int m_id, p_id;
				sscanf(argv[optind], "%x,%x", &m_id, &p_id);
				u = get_by_id(m_id, p_id);
			} else {
				u = get_by_type(opt);
			}

			if (!u) {
				fprintf(stderr, "Unknown device type or id\n");
				exit(1);
			}
			
			break;

		case 2:
			break;

		case 3:
			if (!strcmp("flow", argv[optind]))
				u->flags |=  FLOW_CTL;
			else
				u->flags &= ~FLOW_CTL;
			break;

		case 4:
			u->bdaddr = argv[optind];
			break;
		}
	}

	if (!u) {
		fprintf(stderr, "Unknown device type or id\n");
		exit(1);
	}

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = sig_alarm;
	sigaction(SIGALRM, &sa, NULL);

	/* 5 seconds should be enough for initialization */
	alarm(to);

	uart_fd = init_uart(dev, u, send_break);

	if (uart_fd < 0) {
		perror("Can't initialize device"); 
		exit(1);
	}

	alarm(0);

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);

	sa.sa_handler = sig_hup;
	sigaction(SIGHUP, &sa, NULL);

	p.fd = uart_fd;
	p.events = POLLERR | POLLHUP;//POLLIN;

	if(!flag && port != -1){

		printf("UART initiallized Ok\r\n");

		fcntl(uart_fd, F_SETFL, 0);
		
		fd_path = mkdir(filepath, 0777);
		if(fd_path<0 && EEXIST!=errno){
			printf("error for mkdir(). errno=%d", errno);
			perror( filepath);
			return;
		}

		fd_file = open(filename_test, O_RDWR | O_CREAT |O_TRUNC, DEFFILEMODE);

		if ( fd_file < 0) {
			printf("error for open(). ");
			perror( filename_test);
			return ;
		}

		fd_test_uart = fd_file;
	
		while (-1 == make_socket(&sock_fd, NULL, port)){
			printf("make socket failed!\r\n");
			sleep(5);
		}
		pthread_create(&task, NULL, transfer_loop, "B");

		number = 0;
	}
	
	printf("BCSP: start enter main loop... \n");
	while (!__io_canceled) {
		
		if(flag){	
			
		p.revents = 0;
		err = poll(&p, 1, 500);
		if (err < 0 && errno == EINTR)
			continue;
		if (err>0)
			break;		
					
			
			if(number == 2)
			{
				pthread_create(&hciattach_pid, NULL, hciattach_thread, "A");
			}

			//CSR chip warn reset, exit the loop.
			if(warn_reset)
			{
				__io_canceled = 1;
			}
			else if(exit_flag == 1 && number >= 20)
			{
				__io_canceled = 1;
			}
			number++;
		}
		else if(port != -1){
			printf("%s <--- %s\n", TTYS0_DEV, BT_DEV);
			sleep(2);
			memset(buff, 0, BUFF_LEN);
			read_count = read(uart_fd, buff, BUFF_LEN);			
			printf("read_count = %d, errno = %d\n",read_count, errno);
			if(read_count > 0) {
				write_count = write(sock_fd, buff, read_count);
				printf("%d ---> %d\n", read_count, write_count);
			}
		}
	
	}
	printf("BCSP: exit main loop... \n");

	/* Restore TTY line discipline */
	ld = N_TTY;
	if (ioctl(uart_fd, TIOCSETD, &ld) < 0) {
		perror("Can't restore line discipline");
		exit(1);
	}

	if(flag)
	{
		flag = 0;
		__io_canceled = 0;
		goto CSR;
	}
	if(sock_fd>0)
		close(sock_fd);
	close(uart_fd);
	if(fd_file>0)
		close(fd_file);

	return 0;
}

static void cmd_rstat(int ctl, int hdev, char *opt)
{
	/* Reset HCI device stat counters */
	if (ioctl(ctl, HCIDEVRESTAT, hdev) < 0) {
		fprintf(stderr, "Can't reset stats counters hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_up(int ctl, int hdev, char *opt)
{
        /* Start HCI device */
       	if (ioctl(ctl, HCIDEVUP, hdev) < 0) {
       	        if (errno == EALREADY)
       	                return;
       	        fprintf(stderr, "Can't init device hci%d: %s (%d)\n",
       	                                        hdev, strerror(errno), errno);
	      	exit(1);
	}
}

static void cmd_down(int ctl, int hdev, char *opt)
{
	/* Stop HCI device */
	if (ioctl(ctl, HCIDEVDOWN, hdev) < 0) {
		fprintf(stderr, "Can't down device hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_reset(int ctl, int hdev, char *opt)
{
	/* Reset HCI device */
	cmd_down(ctl, hdev, "down");
	cmd_up(ctl, hdev, "up");
}

static struct {
	char *cmd;
	void (*func)(int ctl, int hdev, char *opt);
	char *opt;
	char *doc;
} command[] = {
	{ "up",		cmd_up,		0,		"Open and initialize HCI device" },
	{ "down",	cmd_down,	0,		"Close HCI device" },
	{ "reset",	cmd_reset,	0,		"Reset HCI device" },
	{ "rstat",	cmd_rstat,	0,		"Reset statistic counters" },
	{ NULL, NULL, 0, NULL }
};

#define BTADDRDIR "/data/tcmd/BT_Address.txt"
#define BTBUFSIZE 256

static char *get_bt_addr()
{
        int fd, len = 0;
        char *buffer;

        buffer = (char *)malloc(BTBUFSIZE);
        fd = open(BTADDRDIR, O_RDONLY);
        if(fd < 0)
        {
                strcpy(buffer, (char *)"FFFFFFFFFFFF");
                return buffer;
        }

        len = read(fd, buffer, BTBUFSIZE);
        close(fd);

        printf("The string len is %d.\n", len);

        if(len < 0)
        {
                return buffer;
        }
        buffer[len] = 0;

        return buffer;
}

static void *hciattach_thread(void *argv)
{
	int ctl, dd;
	char *str = NULL;
	unsigned int btnum;
	struct hci_dev_info di;
	uint8_t bdaddr[8] = {0x04, 0x00, 0x06, 0x05, 0x42, 0x00, 0x32, 0x01}; //bluetooth address 01:32:42:04:05:06

        printf("LIAO: HCI dev up start......\n");

	/* Open HCI socket  */
	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0) {
		perror("Can't open HCI socket.");
		exit(1);
	}

	di.dev_id = 0;

	cmd_up(ctl, di.dev_id, NULL);

	if(strlen((char *)argv) <= 2)
	{
	        printf("LIAO: CSR chip pskey setting......\n");
        	dd = hci_open_dev(di.dev_id);

		str = get_bt_addr();
		sscanf(str, "%02X", &btnum);
		bdaddr[7] = btnum;

		sscanf(str+2, "%02X", &btnum);
		bdaddr[6] = btnum;

		sscanf(str+4, "%02X", &btnum);
		bdaddr[4] = btnum;

		sscanf(str+6, "%02X", &btnum);
		bdaddr[0] = btnum;

		sscanf(str+8, "%02X", &btnum);
		bdaddr[3] = btnum;

		sscanf(str+10, "%02X", &btnum);
		bdaddr[2] = btnum;

		if(str != NULL)
		{
			free(str);
			str = NULL;
		}

        	if(csr_write_pskey_uint16(dd, 0x00, CSR_PSKEY_ANA_FREQ, 0x0008, 0x6590))
                	printf("csr_write_pskey_uint16 SETUP_PSKEY_ANA_FREQ failed.\n");

		////////////////////////////////////////////////////////////////////////////
		// 9.6k		0x0027		19.2k	0x004F		38.4k	0x009D
		// 57.6k	0x00EC		115.2k	0x01D8		230.4k	0x03B0
		// 460.8k	0x075F		921.6k	0x0EBF		1382.4k	0x161E
		////////////////////////////////////////////////////////////////////////////
		if(csr_write_pskey_uint16(dd, 0x01, CSR_PSKEY_UART_BAUDRATE, 0x0008, 0x1D82))
			printf("csr_write_pskey_uint16 SETUP_UART failed.\n");

		// Set BT Address
		if(csr_write_pskey_bdaddr(dd, 0x02, CSR_PSKEY_BDADDR, 0x0008, bdaddr, 8))
        	        printf("csr_write_pskey_uint16 SETUP_BDADDR failed.\n");

	        if(csr_write_pskey_uint16(dd, 0x03, CSR_PSKEY_HOSTIO_MAP_SCO_PCM, 0x0008, 0x0001)) //0x0000 HCI 0x0001 PCM
	                printf("csr_write_pskey_uint16 SETUP_PSKEY_HOSTIO_MAP_SCO_PCM failed.\n");

        	if(csr_write_pskey_uint16(dd, 0x04, CSR_PSKEY_HOSTIO_MAP_SCO_CODEC, 0x0008, 0x0000))
                	printf("csr_write_pskey_uint16 SETUP_PSKEY_HOSTIO_MAP_SCO_CODEC failed.\n");
                
		////////////////////////////////////////////////////
		// 0 CPU_FAST (full rate) 
	        // 1 CPU_SLOW_4M (4 MHz) 
	        // 2 CPU_SLOW_2M (2 MHz) 
	        // 3 CPU_SLOW_1p024M (1.024 MHz)
		////////////////////////////////////////////////////
	        if(csr_write_pskey_uint16(dd, 0x05, CSR_PSKEY_PCM_MIN_CPU_CLOCK, 0x0008, 0x0000))//0x0000 or 0x0003
        	        printf("csr_write_pskey_uint16 SETUP_PSKEY_PCM_MIN_CPU_CLOCK failed.\n");

	        if(csr_write_pskey_uint32(dd, 0x06, CSR_PSKEY_PCM_CONFIG32, 0x0008, 0x08800006))
        	        printf("csr_write_pskey_uint32 SET_PCM_CONFIG32 failed.\n");

	        if(csr_write_varid_valueless(dd, 0x07, CSR_VARID_WARM_RESET))
			printf("csr_write_varid_valueless CSR_VARID_WARM_RESET is failed.\n");

		hci_close_dev(dd);

        	/* Stop HCI device */
	        if (ioctl(ctl, HCIDEVDOWN, di.dev_id) < 0) 
		{
             		exit(1);
        	}
	}

        printf("LIAO: HCI dev up end......\n");
	close(ctl);
	warn_reset = 1;
	pthread_exit(NULL);

	return 0;
}

