/*
 *  HCI UART driver initialization utility for Bluetooth chipset
 *  Borqs for Bluetooth chipset (CSR BC4; CSR BC6; Marvell 8688)
 *  Liao Cheng <cheng.liao@borqs.com>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <termios.h>
#include <time.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <pthread.h>
#include "cutils/properties.h"

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include "csr.h"
#include <hardware_legacy/wifi.h>

#ifndef N_TTY
#define N_TTY	0
#endif

#ifndef N_HCI
#define N_HCI	15
#endif

#define HCIUARTSETPROTO		_IOW('U', 200, int)

#define HCI_UART_H4	0
#define HCI_UART_BCSP	1

#define CSR_BC4		0
#define CSR_BC5		1
#define CSR_BC6		2
#define MRVL_8688	3
#define	BCM_2035	4
#define BCM_2045	5

#define FLOW_CTL	0x0001
#define BTADDRDIR "/data/tcmd/BT_Address.txt"
#define BTADDRDIR_OPL "/opl/nvm/BT_Address.txt"

struct uart_t {
	char *type;
	int  m_id;
	int  p_id;
	int  proto;
	int  flags;
	char *bdaddr;
	int  (*init) (int fd, struct uart_t *u, struct termios *ti);
};

static volatile sig_atomic_t __io_canceled = 0;
static volatile sig_atomic_t load_8688_fw = 0;
static int warn_reset = 0;
static void *hciattach_thread_csrbc4(void *argv);
static void *hciattach_thread_mrvl8688(void *argv);
static int uart_fd = -1;
static int bt_chip_type = 0;
static void cmd_scan(int ctl, int hdev, char *opt);
static void cmd_down(int ctl, int hdev, char *opt);

static int get_chipset_type()
{
	int ret = 0;
	char chipname[PROPERTY_VALUE_MAX];
	property_get("hw.bluetooth.chipset", chipname, NULL);

	if(strcmp(chipname, "mrvl8688") == 0)
		ret = MRVL_8688;
	else if(strcmp(chipname, "csrbc4") == 0)
		ret = CSR_BC4;
	else if(strcmp(chipname, "csrbc6") == 0)
		ret = CSR_BC6;
	else if(strcmp(chipname, "bcm2045") == 0)
		ret = BCM_2045;
	else
		ret = CSR_BC4;

	return ret;
}

static void sig_hup(int sig)
{
}

static void sig_term(int sig)
{
	int ctl = -1;

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0) {
		fprintf(stderr, "sig_term: failed to create ctl\n");
		__io_canceled = 1;

		return;
	}

	if(bt_chip_type == MRVL_8688) {
		cmd_scan(ctl, 0, "noscan");
        cmd_down(ctl, 0, "down");
		system("echo 1 > /sys/class/bluetooth/lpm_ctrl");
	}

	close(ctl);

	__io_canceled = 1;
}


static void unexpected_exit_hciattach(){
	if(load_8688_fw){
            wifi_unload_driver();
            load_8688_fw = 0;
        }
	if(bt_chip_type != MRVL_8688) {
		reset_uart(uart_fd);
	}
	usleep(100000);
	close(uart_fd);
}

static void sig_alarm(int sig)
{
	fprintf(stderr, "Initialization timed out.\n");
	unexpected_exit_hciattach();
	exit(1);
}

static int set_speed(int fd, struct termios *ti)
{
	if(warn_reset)
	{
		cfsetospeed(ti, B1500000);
		cfsetispeed(ti, B1500000);
	}
	else if(bt_chip_type == MRVL_8688)
	{
		cfsetospeed(ti, B38400);
		cfsetispeed(ti, B38400);
	}
	else
	{
		cfsetospeed(ti, B921600);
		cfsetispeed(ti, B921600);
	}

	return tcsetattr(fd, TCSANOW, ti);
}

static int read_check(int fd, void *buf, int count)
{
	int res;

	do {
		res = read(fd, buf, count);
		if (res != -1) {
			buf += res; 
			count -= res;
		}
	} while (count && (errno == 0 || errno == EINTR));

	if (count)
		return -1;

	return 0;
}

/*
 * BCSP specific initialization
 */
int serial_fd;

static void bcsp_tshy_sig_alarm(int sig)
{
	static int retries=0;
	unsigned char bcsp_sync_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xda,0xdc,0xed,0xed,0xc0};
	int len;

	if (retries < 20) {
		retries++;
		len = write(serial_fd, &bcsp_sync_pkt, 10);
		alarm(1);
		return;
	}

	tcflush(serial_fd, TCIOFLUSH);
	fprintf(stderr, "BCSP initialization timed out\n");
	unexpected_exit_hciattach();
	exit(1);
}

static void bcsp_tconf_sig_alarm(int sig)
{
	static int retries=0;
	unsigned char bcsp_conf_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xad,0xef,0xac,0xed,0xc0};
	int len;

	if (retries < 10){
		retries++;
		len = write(serial_fd, &bcsp_conf_pkt, 10);
		alarm(1);
		return;
	}

	tcflush(serial_fd, TCIOFLUSH);
	fprintf(stderr, "BCSP initialization timed out\n");
	unexpected_exit_hciattach();
	exit(1);
}

static int bcsp(int fd, struct uart_t *u, struct termios *ti)
{
	unsigned char byte, bcsph[4], bcspp[4],
		bcsp_sync_resp_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xac,0xaf,0xef,0xee,0xc0},
		bcsp_conf_resp_pkt[10] = {0xc0,0x00,0x41,0x00,0xbe,0xde,0xad,0xd0,0xd0,0xc0},
		bcspsync[4]     = {0xda, 0xdc, 0xed, 0xed},
		bcspsyncresp[4] = {0xac,0xaf,0xef,0xee},
		bcspconf[4]     = {0xad,0xef,0xac,0xed},
		bcspconfresp[4] = {0xde,0xad,0xd0,0xd0};
	struct sigaction sa;
	int len;

	printf("hciattach bcsp for CSR chip...\n");
	if (set_speed(fd, ti) < 0) {
		perror("Can't set default baud rate");
		return -1;
	}

	ti->c_cflag |= PARENB;
	ti->c_cflag &= ~(PARODD);

	if (tcsetattr(fd, TCSANOW, ti) < 0) {
		perror("Can't set port settings");
		return -1;
	}

	alarm(0);

	serial_fd = fd;
	memset(&sa, 0, sizeof(sa));
	sa.sa_flags = SA_NOCLDSTOP;
	sa.sa_handler = bcsp_tshy_sig_alarm;
	sigaction(SIGALRM, &sa, NULL);

	/* State = shy */

	bcsp_tshy_sig_alarm(0);
	while (1) {
		do {
			if (read_check(fd, &byte, 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (byte != 0xC0);

		do {
			if ( read_check(fd, &bcsph[0], 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (bcsph[0] == 0xC0);

		if ( read_check(fd, &bcsph[1], 3) == -1){
			perror("Failed to read");
			return -1;
		}

		if (((bcsph[0] + bcsph[1] + bcsph[2]) & 0xFF) != (unsigned char)~bcsph[3])
			continue;
		if (bcsph[1] != 0x41 || bcsph[2] != 0x00)
			continue;

		if (read_check(fd, &bcspp, 4) == -1){
			perror("Failed to read");
			return -1;
		}

		if (!memcmp(bcspp, bcspsync, 4)) {
			len = write(fd, &bcsp_sync_resp_pkt,10);
		} else if (!memcmp(bcspp, bcspsyncresp, 4))
			break;
	}

	/* State = curious */

	alarm(0);
	sa.sa_handler = bcsp_tconf_sig_alarm;
	sigaction(SIGALRM, &sa, NULL);
	alarm(1);

	while (1) {
		do {
			if (read_check(fd, &byte, 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (byte != 0xC0);

		do {
			if (read_check(fd, &bcsph[0], 1) == -1){
				perror("Failed to read");
				return -1;
			}
		} while (bcsph[0] == 0xC0);

		if (read_check(fd, &bcsph[1], 3) == -1){
			perror("Failed to read");
			return -1;
		}

		if (((bcsph[0] + bcsph[1] + bcsph[2]) & 0xFF) != (unsigned char)~bcsph[3])
			continue;

		if (bcsph[1] != 0x41 || bcsph[2] != 0x00)
			continue;

		if (read_check(fd, &bcspp, 4) == -1){
			perror("Failed to read");
			return -1;
		}

		if (!memcmp(bcspp, bcspsync, 4))
			len = write(fd, &bcsp_sync_resp_pkt, 10);
		else if (!memcmp(bcspp, bcspconf, 4))
			len = write(fd, &bcsp_conf_resp_pkt, 10);
		else if (!memcmp(bcspp, bcspconfresp,  4))
			break;
	}

	return 0;
}

struct uart_t uart[] = {
	{ "any",        0x0000, 0x0000, HCI_UART_H4,	FLOW_CTL, NULL, NULL    },
	{ "bcsp",       0x0000, 0x0000, HCI_UART_BCSP, 	0,        NULL, bcsp    },
	{ NULL, 	0, 	0, 	0, 		0, 	NULL, 	NULL 	}
};

struct uart_t * get_by_id(int m_id, int p_id)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (uart[i].m_id == m_id && uart[i].p_id == p_id)
			return &uart[i];
	}
	return NULL;
}

struct uart_t * get_by_type(char *type)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (!strcmp(uart[i].type, type))
			return &uart[i];
	}
	return NULL;
}

int reset_uart(int fd)
{
	struct termios ti;

	if (tcgetattr(fd, &ti) < 0) {
		printf("Can't get port settings.\n");
		return -1;
	}

	if(bt_chip_type == MRVL_8688)
	{
		cfsetospeed(&ti, B38400);
		cfsetispeed(&ti, B38400);
	}
	else
	{
		cfsetospeed(&ti, B921600);
		cfsetispeed(&ti, B921600);
	}

	return tcsetattr(fd, TCSANOW, &ti);
}

/* Initialize UART driver */
int init_uart(char *dev, struct uart_t *u, int send_break)
{
	struct termios ti;
	int fd, i;

	if(warn_reset && (bt_chip_type == CSR_BC4))
		fd = uart_fd;
	else if(warn_reset && (bt_chip_type == CSR_BC6))
		fd = uart_fd;
	else
		fd = open(dev, O_RDWR | O_NOCTTY);
	
    usleep(100000);

	tcflush(fd, TCIOFLUSH);
	if (fd < 0) {
		perror("Can't open serial port");
		return -1;
	}

	if (tcgetattr(fd, &ti) < 0) {
		perror("Can't get port settings");
		return -1;
	}

	cfmakeraw(&ti);

	ti.c_cflag |= CLOCAL;
	if (u->flags & FLOW_CTL)
		ti.c_cflag |= CRTSCTS;
	else
		ti.c_cflag &= ~CRTSCTS;

	if (tcsetattr(fd, TCSANOW, &ti) < 0) {
		perror("Can't set port settings");
		return -1;
	}

	/* Set initial baudrate */
	if (set_speed(fd, &ti) < 0) {
		perror("Can't set initial baud rate");
		return -1;
	}

	tcflush(fd, TCIOFLUSH);

	if (send_break) {
		tcsendbreak(fd, 0);
		usleep(500000);
	}

	if (u->init && u->init(fd, u, &ti) < 0)
		return -1;

	tcflush(fd, TCIOFLUSH);

	if (set_speed(fd, &ti) < 0) {
		perror("Can't set baud rate");
		return -1;
	}

	i = N_HCI;
	if (ioctl(fd, TIOCSETD, &i) < 0) {
		perror("Can't set line discipline");
		return -1;
	}

	if (ioctl(fd, HCIUARTSETPROTO, u->proto) < 0) {
		perror("Can't set device");
		return -1;
	}

	return fd;
}

#define MRVL_8688_POWER_ON_SYS_PATH "/sys/class/net/eth0"
int checkFwActive()
{
	int i;

	for (i=0; i<50; i++)
	{
		struct stat state;
		int ret = stat(MRVL_8688_POWER_ON_SYS_PATH, &state);
		if (ret == 0)
			return 0;

		usleep(100000);
	}

	return 1;
}

static const char DRIVER_REF_PROP_NAME[]= "wlan.driver.ref";

int main(int argc, char *argv[])
{
	struct uart_t *u = NULL;
	int opt, i, n, ld, err;
	int to = 5;
	int send_break = 0;
	pid_t pid;
	struct sigaction sa;
	struct pollfd p;
	char dev[PATH_MAX];

	int flag = 0;
	pthread_t hciattach_pid;
	int ref;
	char driver_ref[PROPERTY_VALUE_MAX];
	int fd = -1;
	int lock_ret = -1;
	int have_lock = 0;

	bt_chip_type = get_chipset_type();
	printf("Got the bluetooth chipset type is %d.\n", bt_chip_type);

	if(bt_chip_type == MRVL_8688) {
		warn_reset = 1;
	} else {
		warn_reset = 0;
	}

	while ((opt=getopt(argc, argv, "mbnfpt:s")) != EOF) {
		switch(opt) {
		case 'm':
			flag = 1;
			break;

		case 'n':
			break;

		case 'f':
			if(bt_chip_type == MRVL_8688){
				load_8688_fw = 1;
				wifi_load_driver();
				usleep(1500000);
				if(checkFwActive()){
					wifi_unload_driver();
					exit(1);
				}
			}
			break;
		case 'b':
			send_break = 1;
			break;

		case 't':
			to = atoi(optarg);
			break;

		default:
			goto exit_hciattach;
		}
	}

	n = argc - optind;
	if (n < 2) {
		exit(1);
	}
BTRESET:
	for (n = 0; optind < argc; n++, optind++) {
		char *opt;

		opt = argv[optind];

		switch(n) {
		case 0:
			dev[0] = 0;
			if (!strchr(opt, '/'))
				strcpy(dev, "/dev/");
			strcat(dev, opt);
			break;

		case 1:
			if((bt_chip_type == CSR_BC4) || (bt_chip_type == CSR_BC6))
			{
				u = get_by_type("bcsp");
			}
			else if(bt_chip_type == MRVL_8688)
			{
				u = get_by_type("any");
			}

			if (!u) {
				fprintf(stderr, "Unknown device type or id\n");
				goto exit_hciattach;
			}
			
			break;

		case 2:
			break;

		case 3:
			if((bt_chip_type == CSR_BC4) || (bt_chip_type == CSR_BC6))
			{
				u->flags &= ~FLOW_CTL;
			}
			else if(bt_chip_type == MRVL_8688)
			{
				u->flags |=  FLOW_CTL;
			}
			break;

		case 4:
			u->bdaddr = argv[optind];
			break;
		}
	}

	if (!u) {
		fprintf(stderr, "Unknown device type or id\n");
		goto exit_hciattach;
	}

	if(bt_chip_type == MRVL_8688) {	
		fd = open("/tmp/hciattach.lock", O_RDWR | O_CREAT, 0666);
		if(!fd) {
			printf("open file /tmp/hciattach.lock error\n");
		}
		lock_ret = flock(fd, LOCK_EX);
		if(lock_ret == 0) {
			have_lock = 1;
		} else {
			printf("Exec flock LOCK_EX failed.\n");
		}
		system("echo 0 > /sys/class/bluetooth/lpm_ctrl");
		//usleep(500000);
	}

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = sig_alarm;
	sigaction(SIGALRM, &sa, NULL);

	/* 5 seconds should be enough for initialization */
	alarm(to);

	uart_fd = init_uart(dev, u, send_break);

	if (uart_fd < 0) {
		perror("Can't initialize device"); 
		goto exit_hciattach;
	}

	alarm(0);

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);

	sa.sa_handler = sig_hup;
	sigaction(SIGHUP, &sa, NULL);

	p.fd = uart_fd;
	p.events = POLLERR | POLLHUP;

	if(flag)
	{
		if(bt_chip_type == MRVL_8688) {
			pthread_create(&hciattach_pid, NULL, hciattach_thread_mrvl8688, "A");
			flag = 0;
		} else {
			pthread_create(&hciattach_pid, NULL, hciattach_thread_csrbc4, "A");
		}
	}

	printf("BCSP: start enter main loop... \n");
	while (!__io_canceled) {	
		p.revents = 0;
		err = poll(&p, 1, 500);
		if (err < 0 && errno == EINTR)
			continue;
		if (err)
			break;
	}

	printf("BCSP: exit main loop... \n");

	if(0==flag && load_8688_fw)
	{
		wifi_unload_driver();
	}

	/* Restore TTY line discipline */
	ld = N_TTY;
	if (ioctl(uart_fd, TIOCSETD, &ld) < 0) {
		perror("Can't restore line discipline");
		goto exit_hciattach;
	}

	if(flag)
	{
		flag = 0;
		__io_canceled = 0;
		goto BTRESET;
	}
	if (bt_chip_type != MRVL_8688) {
		reset_uart(uart_fd);
	}
	usleep(100000);
	close(uart_fd);

	if(bt_chip_type == MRVL_8688 && have_lock) {
		if(flock(fd, LOCK_UN) < 0) {
			printf("Exec flock LOCK_UN failed.\n");
		}
	}

	return 0;

exit_hciattach:
	unexpected_exit_hciattach();
	if(bt_chip_type == MRVL_8688 && have_lock) {
		if(flock(fd, LOCK_UN) < 0) {
                        printf("Exec flock LOCK_UN failed.\n");
                }
        }
	return 1;
}

static void cmd_rstat(int ctl, int hdev, char *opt)
{
	/* Reset HCI device stat counters */
	printf("Enter cmd_rstat() ...");
	if (ioctl(ctl, HCIDEVRESTAT, hdev) < 0) {
		fprintf(stderr, "Can't reset stats counters hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		unexpected_exit_hciattach();
		exit(1);
	}
}

static void cmd_up(int ctl, int hdev, char *opt)
{
        /* Start HCI device */
	fprintf(stderr, "Enter cmd_up() ...");
       	if (ioctl(ctl, HCIDEVUP, hdev) < 0) {
       	        if (errno == EALREADY)
       	                return;
       	        fprintf(stderr, "Can't init device hci%d: %s (%d)\n",
       	                                        hdev, strerror(errno), errno);
		unexpected_exit_hciattach();
		exit(1);
	}
}

static void cmd_auth(int ctl, int hdev, char *opt)
{
        struct hci_dev_req dr;

        dr.dev_id = hdev;
        if (!strcmp(opt, "auth"))
                dr.dev_opt = AUTH_ENABLED;
        else
                dr.dev_opt = AUTH_DISABLED;

        if (ioctl(ctl, HCISETAUTH, (unsigned long) &dr) < 0) {
                fprintf(stderr, "Can't set auth on hci%d: %s (%d)\n",
                                                hdev, strerror(errno), errno);
		unexpected_exit_hciattach();
                exit(1);
        }
}

static void cmd_down(int ctl, int hdev, char *opt)
{
	/* Stop HCI device */
	fprintf(stderr, "Enter cmd_down() ...");
	if (ioctl(ctl, HCIDEVDOWN, hdev) < 0) {
		fprintf(stderr, "Can't down device hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		unexpected_exit_hciattach();
		exit(1);
	}
}

static void cmd_scan(int ctl, int hdev, char *opt)
{
	struct hci_dev_req dr;
	fprintf(stderr, "Enter cmd_scan() ... %s\n", opt);

	dr.dev_id  = hdev;
	dr.dev_opt = SCAN_DISABLED;
	if (!strcmp(opt, "iscan"))
		dr.dev_opt = SCAN_INQUIRY;
	else if (!strcmp(opt, "pscan"))
		dr.dev_opt = SCAN_PAGE;
	else if (!strcmp(opt, "piscan"))
		dr.dev_opt = SCAN_PAGE | SCAN_INQUIRY;

	if (ioctl(ctl, HCISETSCAN, (unsigned long) &dr) < 0) {
		fprintf(stderr, "Can't set scan mode on hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		unexpected_exit_hciattach();
		exit(1);
	}
}

static void cmd_reset(int ctl, int hdev, char *opt)
{
	/* Reset HCI device */
	cmd_down(ctl, hdev, "down");
	cmd_up(ctl, hdev, "up");
}

static struct {
	char *cmd;
	void (*func)(int ctl, int hdev, char *opt);
	char *opt;
	char *doc;
} command[] = {
	{ "up",		cmd_up,		0,		"Open and initialize HCI device" },
	{ "down",	cmd_down,	0,		"Close HCI device" },
	{ "reset",	cmd_reset,	0,		"Reset HCI device" },
	{ "rstat",	cmd_rstat,	0,		"Reset statistic counters" },
	{ NULL, NULL, 0, NULL }
};

static char *get_bt_addr()
{
        int fd, len = 0;
        char *buffer;

        buffer = (char *)malloc(256);
        fd = open(BTADDRDIR_OPL, O_RDONLY);
        if(fd < 0)
            fd = open(BTADDRDIR, O_RDONLY);

        if(fd < 0)
        {
                strcpy(buffer, (char *)"FFFFFFFFFFFF");
                return buffer;
        }

        len = read(fd, buffer, 256);
        close(fd);

        if(len < 0)
        {
                return buffer;
        }
        buffer[len] = 0;

        return buffer;
}

static void *hciattach_thread_csrbc4(void *argv)
{
	int ctl, dd;
	char *str = NULL;
	unsigned int btnum;
	uint8_t bdaddr[8] = {0x04, 0x00, 0x06, 0x05, 0x42, 0x00, 0x32, 0x01}; //bluetooth address 01:32:42:04:05:06

        printf("LIAO: hciattach_thread_csrbc4 HCI dev up start......\n");

	/* Open HCI socket  */
	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0) {
		perror("Can't open HCI socket.");
		exit(1);
	}

	cmd_up(ctl, 0, NULL);

	if(strlen((char *)argv) <= 2)
	{
	        printf("LIAO: CSR chip pskey setting......\n");
        	dd = hci_open_dev(0);

		str = get_bt_addr();
		sscanf(str, "%02X", &btnum); bdaddr[7] = btnum;
		sscanf(str+2, "%02X", &btnum); bdaddr[6] = btnum;
		sscanf(str+4, "%02X", &btnum); bdaddr[4] = btnum;
		sscanf(str+6, "%02X", &btnum); bdaddr[0] = btnum;
		sscanf(str+8, "%02X", &btnum); bdaddr[3] = btnum;
		sscanf(str+10, "%02X", &btnum); bdaddr[2] = btnum;

		if(str != NULL)
		{
			free(str);
			str = NULL;
		}

        	if(csr_write_pskey_uint16(dd, 0x00, CSR_PSKEY_ANA_FREQ, 0x0008, 0x6590))
                	printf("csr_write_pskey_uint16 SETUP_PSKEY_ANA_FREQ failed.\n");

		////////////////////////////////////////////////////////////////////////////
		// 9.6k		0x0027		19.2k	0x004F		38.4k	0x009D
		// 57.6k	0x00EC		115.2k	0x01D8		230.4k	0x03B0
		// 460.8k	0x075F		921.6k	0x0EBF		1382.4k	0x161E
		////////////////////////////////////////////////////////////////////////////
		if(csr_write_pskey_uint16(dd, 0x01, CSR_PSKEY_UART_BAUDRATE, 0x0008, 0x1D82))
			printf("csr_write_pskey_uint16 SETUP_UART failed.\n");

		// Set BT Address
		if(csr_write_pskey_bdaddr(dd, 0x02, CSR_PSKEY_BDADDR, 0x0008, bdaddr, 8))
        	        printf("csr_write_pskey_uint16 SETUP_BDADDR failed.\n");

	        if(csr_write_pskey_uint16(dd, 0x03, CSR_PSKEY_HOSTIO_MAP_SCO_PCM, 0x0008, 0x0001)) //0x0000 HCI 0x0001 PCM
	                printf("csr_write_pskey_uint16 SETUP_PSKEY_HOSTIO_MAP_SCO_PCM failed.\n");

        	if(csr_write_pskey_uint16(dd, 0x04, CSR_PSKEY_HOSTIO_MAP_SCO_CODEC, 0x0008, 0x0000))
                	printf("csr_write_pskey_uint16 SETUP_PSKEY_HOSTIO_MAP_SCO_CODEC failed.\n");
                
		////////////////////////////////////////////////////
		// 0 CPU_FAST (full rate) 
	        // 1 CPU_SLOW_4M (4 MHz) 
	        // 2 CPU_SLOW_2M (2 MHz) 
	        // 3 CPU_SLOW_1p024M (1.024 MHz)
		////////////////////////////////////////////////////
	        if(csr_write_pskey_uint16(dd, 0x05, CSR_PSKEY_PCM_MIN_CPU_CLOCK, 0x0008, 0x0000))//0x0000 or 0x0003
        	        printf("csr_write_pskey_uint16 SETUP_PSKEY_PCM_MIN_CPU_CLOCK failed.\n");

	        if(csr_write_pskey_uint16(dd, 0x06, CSR_PSKEY_PCM_FORMAT, 0x0008, 0x0060))
        	        printf("csr_write_pskey_uint16 SETUP_CSR_PSKEY_PCM_FORMAT failed.\n");

	        if(csr_write_pskey_uint32(dd, 0x07, CSR_PSKEY_PCM_CONFIG32, 0x0008, 0x08800006))
        	        printf("csr_write_pskey_uint32 SET_PCM_CONFIG32 failed.\n");

	        if(csr_write_varid_valueless(dd, 0x08, CSR_VARID_WARM_RESET))
			printf("csr_write_varid_valueless CSR_VARID_WARM_RESET is failed.\n");

		hci_close_dev(dd);

	        if (ioctl(ctl, HCIDEVDOWN, 0) < 0) 
		{
             		exit(1);
        	}
	}

        printf("LIAO: HCI dev up end......\n");
	close(ctl);
	warn_reset = 1;
	__io_canceled = 1;
	pthread_exit(NULL);

	return 0;
}

static void *hciattach_thread_mrvl8688(void *argv)
{
	int dd, ctl;
	unsigned char cp[254];

	usleep(100000);
	printf("LIAO: hciattach_thread_mrvl8688 HCI dev up start......\n");
	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0) {
		perror("Can't open HCI socket.");
		unexpected_exit_hciattach();
		exit(1);
	}
	//sleep(1);

	printf(" cmd_up() ...");
	cmd_up(ctl, 0, NULL);
	//cmd_scan(ctl, 0, "pscan");
	cmd_auth(ctl, 0, "auth");

	dd = hci_open_dev(0);
	
	memset(&cp, 0, sizeof(cp));
	cp[0] = 0x04;

	if (hci_send_cmd(dd, 0x3f, 0x0007, 1, cp) < 0) {
		printf("Send Command HCI_CMD_MARVELL_WRITE_PCM_SETTINGS failed.\n");
	}
	usleep(50000);

	hci_close_dev(dd);

	close(ctl);
	pthread_exit(NULL);

	return 0;
}
