/*
 *  HCI UART driver initialization utility for Bluetooth chipset
 *  Borqs for Bluetooth chipset (CSR BC4; CSR BC6; Marvell 8688)
 *  Liao Cheng <cheng.liao@borqs.com>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <termios.h>
#include <time.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <pthread.h>
#include "cutils/properties.h"

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>

#include "csr.h"
#include <hardware_legacy/wifi.h>

#ifndef N_TTY
#define N_TTY	0
#endif

#ifndef N_HCI
#define N_HCI	15
#endif

#define HCIUARTSETPROTO		_IOW('U', 200, int)

#define HCI_UART_H4	0
#define HCI_UART_BCSP	1

#define CSR_BC4		0
#define CSR_BC5		1
#define CSR_BC6		2
#define MRVL_8688	3
#define	BCM_2035	4
#define BCM_2045	5

#define FLOW_CTL	0x0001
#define BTADDRDIR "/data/tcmd/BT_Address.txt"

struct uart_t {
	char *type;
	int  m_id;
	int  p_id;
	int  proto;
	int  flags;
	char *bdaddr;
	int  (*init) (int fd, struct uart_t *u, struct termios *ti);
};

static volatile sig_atomic_t __io_canceled = 0;
static volatile sig_atomic_t load_8688_fw = 0;
static void *hciattach_thread_mrvl8688(void *argv);
static int uart_fd = -1;
static int bt_chip_type = 0;

static int get_chipset_type()
{
	int ret = 0;
	char chipname[1024];
	property_get("hw.bluetooth.chipset", chipname, NULL);

	if(strcmp(chipname, "mrvl8688") == 0)
		ret = MRVL_8688;
	else if(strcmp(chipname, "csrbc4") == 0)
		ret = CSR_BC4;
	else if(strcmp(chipname, "csrbc6") == 0)
		ret = CSR_BC6;
	else if(strcmp(chipname, "bcm2045") == 0)
		ret = BCM_2045;
	else
		ret = CSR_BC4;

	return ret;
}

static void sig_hup(int sig)
{
}

static void sig_term(int sig)
{
	__io_canceled = 1;
}

static void sig_alarm(int sig)
{
	fprintf(stderr, "Initialization timed out.\n");
	exit(1);
}

static int set_speed(int fd, struct termios *ti)
{
	cfsetospeed(ti, B38400);
	cfsetispeed(ti, B38400);

	return tcsetattr(fd, TCSANOW, ti);
}

static int read_check(int fd, void *buf, int count)
{
	int res;

	do {
		res = read(fd, buf, count);
		if (res != -1) {
			buf += res; 
			count -= res;
		}
	} while (count && (errno == 0 || errno == EINTR));

	if (count)
		return -1;

	return 0;
}
struct uart_t uart[] = {
	{ "any",        0x0000, 0x0000, HCI_UART_H4,	FLOW_CTL, NULL, NULL    },
	{ NULL, 	0, 	0, 	0, 		0, 	NULL, 	NULL 	}
};

struct uart_t * get_by_id(int m_id, int p_id)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (uart[i].m_id == m_id && uart[i].p_id == p_id)
			return &uart[i];
	}
	return NULL;
}

struct uart_t * get_by_type(char *type)
{
	int i;
	for (i = 0; uart[i].type; i++) {
		if (!strcmp(uart[i].type, type))
			return &uart[i];
	}
	return NULL;
}

int reset_uart(int fd)
{
	struct termios ti;

	if (tcgetattr(fd, &ti) < 0) {
		printf("Can't get port settings.\n");
		return -1;
	}

	cfsetospeed(&ti, B38400);
	cfsetispeed(&ti, B38400);

	return tcsetattr(fd, TCSANOW, &ti);
}

/* Initialize UART driver */
int init_uart(char *dev, struct uart_t *u, int send_break)
{
	struct termios ti;
	int fd, i;

	fd = open(dev, O_RDWR | O_NOCTTY);
	reset_uart(fd);
	//system("echo 0 > /sys/class/bluetooth/lpm_ctrl");
	close(fd);

    usleep(100000);
	fd = open(dev, O_RDWR | O_NOCTTY);

	tcflush(fd, TCIOFLUSH);
	if (fd < 0) {
		perror("Can't open serial port");
		return -1;
	}

	if (tcgetattr(fd, &ti) < 0) {
		perror("Can't get port settings");
		return -1;
	}

	cfmakeraw(&ti);

	ti.c_cflag |= CLOCAL;
	if (u->flags & FLOW_CTL)
		ti.c_cflag |= CRTSCTS;
	else
		ti.c_cflag &= ~CRTSCTS;

	if (tcsetattr(fd, TCSANOW, &ti) < 0) {
		perror("Can't set port settings");
		return -1;
	}

	/* Set initial baudrate */
	if (set_speed(fd, &ti) < 0) {
		perror("Can't set initial baud rate");
		return -1;
	}

	tcflush(fd, TCIOFLUSH);

	if (send_break) {
		tcsendbreak(fd, 0);
		usleep(500000);
	}

	if (u->init && u->init(fd, u, &ti) < 0)
		return -1;

	tcflush(fd, TCIOFLUSH);

	if (set_speed(fd, &ti) < 0) {
		perror("Can't set baud rate");
		return -1;
	}

	i = N_HCI;
	if (ioctl(fd, TIOCSETD, &i) < 0) {
		perror("Can't set line discipline");
		return -1;
	}

	if (ioctl(fd, HCIUARTSETPROTO, u->proto) < 0) {
		perror("Can't set device");
		return -1;
	}

	return fd;
}

#define MRVL_8688_POWER_ON_SYS_PATH "/sys/class/net/eth0"
int checkFwActive()
{
	int i;

	for (i=0; i<50; i++)
	{
		struct stat state;
		int ret = stat(MRVL_8688_POWER_ON_SYS_PATH, &state);
		if (ret == 0)
			return 0;

		usleep(100000);
	}

	return 1;
}

int main(int argc, char *argv[])
{
	struct uart_t *u = NULL;
	int opt, i, n, ld, err;
	int to = 5;
	int send_break = 0;
	pid_t pid;
	struct sigaction sa;
	struct pollfd p;
	char dev[PATH_MAX];

	pthread_t hciattach_pid;
	int fd = -1;
	int lock_ret = -1;
	int have_lock = 0;

	bt_chip_type = get_chipset_type();
	printf("Got the bluetooth chipset type is %d.\n", bt_chip_type);

	fd = open("/tmp/hciattach.lock", O_RDWR | O_CREAT, 0666);
	if(!fd) {
		printf("open file /tmp/hciattach.lock error\n");
	}
	lock_ret = flock(fd, LOCK_EX);
	if(lock_ret == 0) {
		have_lock = 1;
	} else {
		printf("Exec flock LOCK_EX failed.\n");
	}

	system("echo 0 > /sys/class/bluetooth/lpm_ctrl");

	while ((opt=getopt(argc, argv, "mbnfpt:s")) != EOF) {
		switch(opt) {
		case 'm':
			//flag = 1;
			break;

		case 'n':
			break;

		case 'f':
			break;

		case 'b':
			send_break = 1;
			break;

		case 't':
			to = atoi(optarg);
			break;

		default:
			goto exit_hciattach;
		}
	}

	n = argc - optind;
	if (n < 2) {
		exit(1);
	}

	for (n = 0; optind < argc; n++, optind++) {
		char *opt;

		opt = argv[optind];

		switch(n) {
		case 0:
			dev[0] = 0;
			if (!strchr(opt, '/'))
				strcpy(dev, "/dev/");
			strcat(dev, opt);
			break;

		case 1:
			u = get_by_type("any");
			if (!u) {
				fprintf(stderr, "Unknown device type or id\n");
				goto exit_hciattach;
			}
			
			break;

		case 2:
			break;

		case 3:
			u->flags |=  FLOW_CTL;
			break;

		case 4:
			u->bdaddr = argv[optind];
			break;
		}
	}

	if (!u) {
		fprintf(stderr, "Unknown device type or id\n");
		goto exit_hciattach;
	}

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = sig_alarm;
	sigaction(SIGALRM, &sa, NULL);

	/* 5 seconds should be enough for initialization */
	alarm(to);

	uart_fd = init_uart(dev, u, send_break);

	if (uart_fd < 0) {
		perror("Can't initialize device"); 
		goto exit_hciattach;
	}

	alarm(0);

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags   = SA_NOCLDSTOP;
	sa.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGPIPE, &sa, NULL);

	sa.sa_handler = sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);

	sa.sa_handler = sig_hup;
	sigaction(SIGHUP, &sa, NULL);

	p.fd = uart_fd;
	p.events = POLLERR | POLLHUP;

	pthread_create(&hciattach_pid, NULL, hciattach_thread_mrvl8688, "A");

	printf("BCSP: start enter main loop... \n");
	while (!__io_canceled) {	
		p.revents = 0;
		err = poll(&p, 1, 500);
		if (err < 0 && errno == EINTR)
			continue;
		if (err)
			break;
	}

	printf("BCSP: exit main loop... \n");

	/* Restore TTY line discipline */
	ld = N_TTY;
	if (ioctl(uart_fd, TIOCSETD, &ld) < 0) {
		perror("Can't restore line discipline");
		goto exit_hciattach;
	}

	close(uart_fd);
	usleep(100000);

	if(have_lock) {
		if(flock(fd, LOCK_UN) < 0) {
			printf("Exec flock LOCK_UN failed.\n");
		}
	}

	system("echo 1 > /sys/class/bluetooth/lpm_ctrl");

	return 0;

exit_hciattach:
	if(have_lock) {
		if(flock(fd, LOCK_UN) < 0) {
                        printf("Exec flock LOCK_UN failed.\n");
                }
        }

	system("echo 1 > /sys/class/bluetooth/lpm_ctrl");

	return 1;
}

static void cmd_rstat(int ctl, int hdev, char *opt)
{
	/* Reset HCI device stat counters */
	printf("Enter cmd_rstat() ...");
	if (ioctl(ctl, HCIDEVRESTAT, hdev) < 0) {
		fprintf(stderr, "Can't reset stats counters hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_up(int ctl, int hdev, char *opt)
{
        /* Start HCI device */
	fprintf(stderr, "Enter cmd_up() ...");
       	if (ioctl(ctl, HCIDEVUP, hdev) < 0) {
       	        if (errno == EALREADY)
       	                return;
       	        fprintf(stderr, "Can't init device hci%d: %s (%d)\n",
       	                                        hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_down(int ctl, int hdev, char *opt)
{
	/* Stop HCI device */
	fprintf(stderr, "Enter cmd_down() ...");
	if (ioctl(ctl, HCIDEVDOWN, hdev) < 0) {
		fprintf(stderr, "Can't down device hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_scan(int ctl, int hdev, char *opt)
{
	struct hci_dev_req dr;
	fprintf(stderr, "Enter cmd_scan() ... %s\n", opt);

	dr.dev_id  = hdev;
	dr.dev_opt = SCAN_DISABLED;
	if (!strcmp(opt, "iscan"))
		dr.dev_opt = SCAN_INQUIRY;
	else if (!strcmp(opt, "pscan"))
		dr.dev_opt = SCAN_PAGE;
	else if (!strcmp(opt, "piscan"))
		dr.dev_opt = SCAN_PAGE | SCAN_INQUIRY;

	if (ioctl(ctl, HCISETSCAN, (unsigned long) &dr) < 0) {
		fprintf(stderr, "Can't set scan mode on hci%d: %s (%d)\n",
						hdev, strerror(errno), errno);
		exit(1);
	}
}

static void cmd_reset(int ctl, int hdev, char *opt)
{
	/* Reset HCI device */
	cmd_down(ctl, hdev, "down");
	cmd_up(ctl, hdev, "up");
}

static struct {
	char *cmd;
	void (*func)(int ctl, int hdev, char *opt);
	char *opt;
	char *doc;
} command[] = {
	{ "up",		cmd_up,		0,		"Open and initialize HCI device" },
	{ "down",	cmd_down,	0,		"Close HCI device" },
	{ "reset",	cmd_reset,	0,		"Reset HCI device" },
	{ "rstat",	cmd_rstat,	0,		"Reset statistic counters" },
	{ NULL, NULL, 0, NULL }
};

static char *get_bt_addr()
{
        int fd, len = 0;
        char *buffer;

        buffer = (char *)malloc(256);
        fd = open(BTADDRDIR, O_RDONLY);
        if(fd < 0)
        {
                strcpy(buffer, (char *)"FFFFFFFFFFFF");
                return buffer;
        }

        len = read(fd, buffer, 256);
        close(fd);

        if(len < 0)
        {
                return buffer;
        }
        buffer[len] = 0;

        return buffer;
}

static void *hciattach_thread_mrvl8688(void *argv)
{
	int dd, ctl;
	unsigned char cp[254];

	usleep(100000);
	printf("LIAO: hciattach_thread_mrvl8688 HCI dev up start......\n");
	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI)) < 0) {
		perror("Can't open HCI socket.");
		exit(1);
	}
	//sleep(1);

	cmd_up(ctl, 0, NULL);
	cmd_scan(ctl, 0, "piscan");

	dd = hci_open_dev(0);
	
/*	memset(&cp, 0, sizeof(cp));
	cp[0] = 0x04;

	if (hci_send_cmd(dd, 0x3f, 0x0007, 1, cp) < 0) {
		printf("Send Command HCI_CMD_MARVELL_WRITE_PCM_SETTINGS failed.\n");
	}
	usleep(50000);
*/
	memset(&cp, 0, sizeof(cp));
	cp[0] = 0x00;
	cp[1] = 0x20;
	cp[2] = 0x1c;
	cp[3] = 0x00;

	if (hci_send_cmd(dd, 0x3f, 0x0009, 4, cp) < 0) {
		printf("Send Command HCI_CMD_MARVELL_UART_BAUD failed.\n");
	}
	usleep(50000);

	printf("LIAO: HCI dev up end......\n");
	hci_close_dev(dd);
	
	close(ctl);
	__io_canceled = 1;
	pthread_exit(NULL);

	return 0;
}
